package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.BufferUploader
import com.mojang.blaze3d.vertex.DefaultVertexFormat
import com.mojang.blaze3d.vertex.PoseStack
import com.mojang.blaze3d.vertex.Tesselator
import com.mojang.blaze3d.vertex.VertexFormat
import net.minecraft.client.gui.Font
import net.minecraft.client.gui.GuiComponent
import net.minecraft.client.gui.screens.inventory.tooltip.ClientTooltipComponent
import net.minecraft.client.gui.screens.inventory.tooltip.TooltipRenderUtil
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.client.renderer.MultiBufferSource
import net.minecraft.client.renderer.texture.TextureAtlasSprite
import net.minecraft.network.chat.Component
import net.minecraft.resources.ResourceLocation
import net.minecraft.util.FormattedCharSequence
import net.minecraft.world.item.ItemStack
import net.minecraftforge.client.ForgeHooksClient
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.core.math.RGBAColor
import kotlin.math.PI
import kotlin.math.roundToInt

// polyfill class for 1.19.4 and older
class MGUIGraphics(val pose: PoseStack) {
	val bufferSource: MultiBufferSource.BufferSource get() = minecraft.renderBuffers().crumblingBufferSource()
	val width get() = minecraft.window.guiScaledWidth
	val height get() = minecraft.window.guiScaledHeight
	val font: Font get() = minecraft.font

	fun setColor(red: Float, green: Float, blue: Float, alpha: Float) {
		RenderSystem.setShaderColor(red, green, blue, alpha)
	}

	fun renderFakeItem(poseStack: PoseStack, itemStack: ItemStack, x: Int, y: Int) {
		val global = RenderSystem.getModelViewStack()
		global.pushPose()
		global.last().pose().mul(poseStack.last().pose())
		global.last().normal().mul(poseStack.last().normal())
		RenderSystem.applyModelViewMatrix()
		minecraft.itemRenderer.renderGuiItem(itemStack, x, y)
		global.popPose()
		RenderSystem.applyModelViewMatrix()
	}

	fun drawLine(
		startX: Float,
		startY: Float,
		endX: Float,
		endY: Float,
		width: Float,
		z: Float = 0f,
		color: RGBAColor = RGBAColor.WHITE
	) {
		drawLine(pose.last().pose(), startX, startY, endX, endY, width, z, color)
	}

	fun renderRect(
		x: Float,
		y: Float,
		width: Float,
		height: Float,
		z: Float = 0f,
		color: RGBAColor = RGBAColor.WHITE
	) {
		renderRect(pose.last().pose(), x, y, width, height, z, color)
	}

	fun renderTexturedRect(
		x: Float,
		y: Float,
		width: Float,
		height: Float,
		z: Float = 0f,
		uv: IUVCoords = defaultUV,
		uvWinding: UVWindingOrder = UVWindingOrder.NORMAL,
		color: RGBAColor = RGBAColor.WHITE,
		texture: ResourceLocation? = null
	) {
		renderTexturedRect(pose.last().pose(), x, y, width, height, z, uv, uvWinding, color, texture)
	}

	fun renderCheckerboard(
		x: Float,
		y: Float,
		width: Float,
		height: Float,
		z: Float = 0f,
		color: RGBAColor = RGBAColor.WHITE
	) {
		renderCheckerboard(pose.last().pose(), x, y, width, height, z, color)
	}

	fun renderComponentTooltip(font: Font, lines: MutableList<Component>, x: Int, y: Int, itemStack: ItemStack = ItemStack.EMPTY) {
		if (lines.isNotEmpty()) {
			val mapped = lines.map { ClientTooltipComponent.create(it.visualOrderText) }
			val preEvent = ForgeHooksClient.onRenderTooltipPre(itemStack, pose, x, y, width, height, mapped, font, font)

			if (preEvent.isCanceled) return
			var totalWidth = 0
			var totalHeight = if (lines.size == 1) -2 else 0

			for (line in mapped) {
				val k = line.getWidth(preEvent.font)
				if (k > totalWidth) totalWidth = k
				totalHeight += line.height
			}

			@Suppress("NAME_SHADOWING")
			var x = x + 12
			@Suppress("NAME_SHADOWING")
			var y = y - 12

			if (x + totalWidth >= minecraft.window.guiScaledWidth)
				x = (x - 24 - totalWidth).coerceAtLeast(4)

			if (y + totalHeight + 3 >= minecraft.window.guiScaledHeight)
				y = minecraft.window.guiScaledHeight - totalHeight - 3

			pose.pushPose()
			val tesselator = Tesselator.getInstance()
			val bufferbuilder = tesselator.builder
			RenderSystem.setShader { GameRenderer.getPositionColorShader() }
			bufferbuilder.begin(VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION_COLOR)
			val matrix4f = pose.last().pose()
			TooltipRenderUtil.renderTooltipBackground(GuiComponent::fillGradient, matrix4f, bufferbuilder, x, y, totalWidth, totalHeight, 400)
			RenderSystem.enableDepthTest()
			RenderSystem.disableTexture()
			RenderSystem.enableBlend()
			RenderSystem.defaultBlendFunc()
			BufferUploader.drawWithShader(bufferbuilder.end())
			RenderSystem.disableBlend()
			RenderSystem.enableTexture()

			pose.translate(0.0f, 0.0f, 400.0f)
			var yCurrent = y

			for (i in mapped.indices) {
				val line = mapped[i]
				line.renderText(preEvent.font, x, yCurrent, matrix4f, bufferSource)
				yCurrent += line.height + if (i == 0) 2 else 0
			}

			bufferSource.endBatch()

			yCurrent = y

			for (i in mapped.indices) {
				val line = mapped[i]
				line.renderImage(preEvent.font, x, yCurrent, pose, minecraft.itemRenderer, 0 /* blit offset, i love you, go commit self-murder */)
				/* because your existence servers ZERO FUCKING PURPOSE */
				/* Z-BUFFER IN MY GUI PIECE OF SHIT */
				yCurrent += line.height + if (i == 0) 2 else 0
			}

			pose.popPose()
		}
	}

	fun flush() {
		bufferSource.endBatch()
	}

	fun draw(
		text: String,
		x: Float = 0f,
		y: Float = 0f,
		gravity: RenderGravity = RenderGravity.TOP_LEFT,
		scale: Float = 1f,
		color: RGBAColor = RGBAColor.WHITE,
		drawShadow: Boolean = false,
		displayMode: Boolean = false,
		packedLightCoords: Int = 15728880,
		effectColor: Int = 0,
		shadowColor: RGBAColor = RGBAColor.BLACK,
		shadowX: Float = 1f,
		shadowY: Float = 1f,
		shadowZ: Float = -0.1f,
		customShadow: Boolean = false,
		rounding: GravityRounding = if (scale == 1f) GravityRounding.DEFAULT else GravityRounding.NO,
		drawOutline: Boolean = false,
		outlineColor: RGBAColor = RGBAColor.BLACK,
		outlineZ: Float = -0.1f,
		font: Font = this.font,
	): Float {
		val width = font.draw(
			poseStack = pose,
			text = text,
			x = x,
			y = y,
			buffer = bufferSource,
			gravity = gravity,
			scale = scale,
			color = color,
			drawShadow = drawShadow,
			displayMode = displayMode,
			packedLightCoords = packedLightCoords,
			effectColor = effectColor,
			shadowColor = shadowColor,
			shadowX = shadowX,
			shadowY = shadowY,
			shadowZ = shadowZ,
			customShadow = customShadow,
			rounding = rounding,
			drawOutline = drawOutline,
			outlineColor = outlineColor,
			outlineZ = outlineZ,
		)

		flush()
		return width
	}

	fun draw(
		text: Component,
		x: Float = 0f,
		y: Float = 0f,
		gravity: RenderGravity = RenderGravity.TOP_LEFT,
		scale: Float = 1f,
		color: RGBAColor = RGBAColor.WHITE,
		drawShadow: Boolean = false,
		displayMode: Boolean = false,
		packedLightCoords: Int = 15728880,
		effectColor: Int = 0,
		shadowColor: RGBAColor = RGBAColor.BLACK,
		shadowX: Float = 1f,
		shadowY: Float = 1f,
		shadowZ: Float = -0.1f,
		customShadow: Boolean = false,
		rounding: GravityRounding = if (scale == 1f) GravityRounding.DEFAULT else GravityRounding.NO,
		drawOutline: Boolean = false,
		outlineColor: RGBAColor = RGBAColor.BLACK,
		outlineZ: Float = -0.1f,
		font: Font = this.font,
	): Float {
		val width = font.draw(
			poseStack = pose,
			text = text,
			x = x,
			y = y,
			buffer = bufferSource,
			gravity = gravity,
			scale = scale,
			color = color,
			drawShadow = drawShadow,
			displayMode = displayMode,
			packedLightCoords = packedLightCoords,
			effectColor = effectColor,
			shadowColor = shadowColor,
			shadowX = shadowX,
			shadowY = shadowY,
			shadowZ = shadowZ,
			customShadow = customShadow,
			rounding = rounding,
			drawOutline = drawOutline,
			outlineColor = outlineColor,
			outlineZ = outlineZ,
		)

		flush()
		return width
	}

	fun draw(
		text: FormattedCharSequence,
		x: Float = 0f,
		y: Float = 0f,
		gravity: RenderGravity = RenderGravity.TOP_LEFT,
		scale: Float = 1f,
		color: RGBAColor = RGBAColor.WHITE,
		drawShadow: Boolean = false,
		displayMode: Boolean = false,
		packedLightCoords: Int = 15728880,
		effectColor: Int = 0,
		shadowColor: RGBAColor = RGBAColor.BLACK,
		shadowX: Float = 1f,
		shadowY: Float = 1f,
		shadowZ: Float = -0.1f,
		customShadow: Boolean = false,
		rounding: GravityRounding = if (scale == 1f) GravityRounding.DEFAULT else GravityRounding.NO,
		drawOutline: Boolean = false,
		outlineColor: RGBAColor = RGBAColor.BLACK,
		outlineZ: Float = -0.1f,
		font: Font = this.font,
	): Float {
		val width = font.draw(
			poseStack = pose,
			text = text,
			x = x,
			y = y,
			buffer = bufferSource,
			gravity = gravity,
			scale = scale,
			color = color,
			drawShadow = drawShadow,
			displayMode = displayMode,
			packedLightCoords = packedLightCoords,
			effectColor = effectColor,
			shadowColor = shadowColor,
			shadowX = shadowX,
			shadowY = shadowY,
			shadowZ = shadowZ,
			customShadow = customShadow,
			rounding = rounding,
			drawOutline = drawOutline,
			outlineColor = outlineColor,
			outlineZ = outlineZ,
		)

		flush()
		return width
	}

	fun renderSprite(
		sprite: TextureAtlasSprite, x: Float, y: Float,
		width: Float = sprite.contents().width().toFloat(),
		height: Float = sprite.contents().height().toFloat(),
		color: RGBAColor = RGBAColor.WHITE
	) {
		renderTexturedRect(
			x, y, width, height,
			uv = UVCoords(sprite.u0, sprite.v0, sprite.u1, sprite.v1),
			texture = sprite.atlasLocation(),
			color = color
		)
	}

	fun drawArc(
		x: Float,
		y: Float,
		outerRadius: Float,
		innerRadius: Float = 0f,
		startDegree: Double = 0.0,
		endDegree: Double = PI * 2.0,
		steps: Int = (outerRadius * (endDegree - startDegree) * 4.0).roundToInt().coerceAtLeast(12),
		alignAtCenter: Boolean = true
	) = drawArc(pose, x, y, outerRadius, innerRadius, startDegree, endDegree, steps, alignAtCenter)

	companion object {
		private val defaultUV = UVCoords(0f, 0f, 1f, 1f)
	}
}
